# MIT License
#
# Copyright (c) 2020 - Present nxtlo
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implementation for a Bungie a Profile."""

from __future__ import annotations

__all__: typing.Sequence[str] = ("Profile",)

import abc
import datetime
import logging
import typing

import attr

from aiobungie.internal import Time, enums, impl

from .character import Character

log: typing.Final[logging.Logger] = logging.getLogger(__name__)


@attr.s(kw_only=True, hash=True, weakref_slot=False, slots=True, init=True, eq=True)
class ProfileComponentImpl(abc.ABC):
    @property
    @abc.abstractmethod
    def app(self) -> impl.RESTful:
        """A Client that we may use for making other requests."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Profile's name"""

    @property
    @abc.abstractmethod
    def type(self) -> enums.MembershipType:
        """Profile's membership type."""

    @property
    @abc.abstractmethod
    def last_played(self) -> datetime.datetime:
        """The profile user's last played date time."""

    @property
    @abc.abstractmethod
    def is_public(self) -> bool:
        """Profile's privacy status."""

    @property
    @abc.abstractmethod
    def character_ids(self) -> typing.List[int]:
        """A list of the profile's character ids."""

    @property
    @abc.abstractmethod
    def id(self) -> int:
        """The profile's id."""

    async def titan(self) -> Character:
        """Returns the titan character of the profile owner."""
        # We're ignoring the types for the character since
        char: Character = await self.app.rest.fetch_character(
            int(self.id), self.type, enums.Class.TITAN
        )
        return Character(
            id=char.id,
            class_type=char.class_type,
            emblem=char.emblem,
            emblem_hash=char.emblem_hash,
            emblem_icon=char.emblem_icon,
            light=char.light,
            last_played=char.last_played,
            level=char.level,
            member_id=char.member_id,
            member_type=char.member_type,
            gender=char.gender,
            race=char.race,
            total_played_time=char.total_played_time,
            title_hash=char.title_hash,
            stats=char.stats,
        )

    async def hunter(self) -> Character:
        """Returns the hunter character of the profile owner."""
        char: Character = await self.app.rest.fetch_character(
            self.id, self.type, enums.Class.HUNTER
        )
        return Character(
            id=char.id,
            class_type=char.class_type,
            emblem=char.emblem,
            emblem_hash=char.emblem_hash,
            emblem_icon=char.emblem_icon,
            light=char.light,
            last_played=char.last_played,
            level=char.level,
            member_id=char.member_id,
            member_type=char.member_type,
            gender=char.gender,
            race=char.race,
            total_played_time=char.total_played_time,
            title_hash=char.title_hash,
            stats=char.stats,
        )

    async def warlock(self) -> Character:
        """Returns the Warlock character of the profile owner."""
        char: Character = await self.app.rest.fetch_character(
            self.id, self.type, enums.Class.WARLOCK
        )
        return Character(
            id=char.id,
            class_type=char.class_type,
            emblem=char.emblem,
            emblem_hash=char.emblem_hash,
            emblem_icon=char.emblem_icon,
            light=char.light,
            last_played=char.last_played,
            level=char.level,
            member_id=char.member_id,
            member_type=char.member_type,
            gender=char.gender,
            race=char.race,
            total_played_time=char.total_played_time,
            title_hash=char.title_hash,
            stats=char.stats,
        )


@attr.s(kw_only=True, hash=True, weakref_slot=False, slots=True, init=True, eq=True)
class Profile(ProfileComponentImpl):
    """Represents a Bungie member Profile.

    A bungie profile have components, each component has different data,
    for an example, The profile component returns the a `aiobungie.objects.Profile`
    object which's this class, the character component returns `aiobungie.objects.Character` object, etc.
    See `aiobungie.internal.enums.Component` to see the current available components.

    Attributes
    ----------
    id: `builtins.int`
            Profile's id
    name: `builtins.str`
            Profile's name
    type: `aiobungie.internal.enums.MembershipType`
            The profile's membership type.
    last_played: `datetime.datetime`
            The profile owner's last played date in UTC
    character_ids: `typing.List[builtins.int]`
            A list of the profile's character ids.
    power_cap: `builtins.int`
            The profile's current season power cap.
    """

    id: int = attr.ib(repr=True, hash=True, eq=False)
    """Profile's id"""

    app: impl.RESTful = attr.ib(repr=False, hash=False, eq=False)

    name: str = attr.ib(repr=True, hash=False, eq=False)
    """Profile's name."""

    type: enums.MembershipType = attr.ib(repr=True, hash=False, eq=False)
    """Profile's type."""

    is_public: bool = attr.ib(repr=True, hash=False, eq=False)
    """Profile's privacy status."""

    last_played: datetime.datetime = attr.ib(repr=True, hash=False, eq=False)
    """Profile's last played Destiny 2 played date."""

    character_ids: typing.List[int] = attr.ib(repr=False, hash=False, eq=False)
    """A list of the profile's character ids."""

    power_cap: int = attr.ib(repr=False, hash=False, eq=False)
    """The profile's current seaspn power cap."""

    # Components

    def __str__(self) -> str:
        return self.name

    def __int__(self) -> int:
        return int(self.id)

    @property
    def as_dict(self) -> typing.Dict[str, typing.Any]:
        """Returns a dict object of the profile,
        This function is useful if you're binding to other REST apis.
        """
        return dict(
            id=self.id,
            name=self.name,
            type=str(self.type),
            is_public=self.is_public,
            last_played=self.last_played,
            character_ids=self.character_ids,
            power_cap=self.power_cap,
        )

    @property
    def human_timedelta(self) -> str:
        """Returns last_played attr but in human delta date."""
        return Time.human_timedelta(self.last_played)
